@@ -57,7 +57,7 @@ extern "C" {
 #  define MRN_HANDLER_HAVE_HA_INDEX_NEXT_SAME 1
 #endif
 
-#if (MYSQL_VERSION_ID >= 50603) || \
+#if (MYSQL_VERSION_ID >= 50604) || \
     (defined(MRN_MARIADB_P) && MYSQL_VERSION_ID >= 50302)
 #  define MRN_HANDLER_HAVE_HA_CLOSE 1
 #  define MRN_HANDLER_HAVE_MULTI_RANGE_READ 1
@@ -85,6 +85,27 @@ extern "C" {
 #  define MRN_HA_ROWS_FORMAT "lu"
 #endif
 
+#if MYSQL_VERSION_ID >= 50604
+#  define MRN_HAVE_MYSQL_TYPE_TIMESTAMP2
+#  define MRN_HAVE_MYSQL_TYPE_DATETIME2
+#  define MRN_HAVE_MYSQL_TYPE_TIME2
+#endif
+
+#if (MYSQL_VERSION_ID < 50519) || \
+    (50600 <= MYSQL_VERSION_ID && MYSQL_VERSION_ID < 50604)
+#  define MRN_NEED_FREE_STRING_MEMALLOC_PLUGIN_VAR
+#endif
+
+#if MYSQL_VERSION_ID >= 50500
+#  define MRN_HAVE_HA_EXTRA_ADD_CHILDREN_LIST
+#  define MRN_HAVE_HA_EXTRA_IS_ATTACHED_CHILDREN
+#endif
+
+#ifdef MRN_MARIADB_P
+#  define MRN_HAVE_HA_EXTRA_DETACH_CHILD
+#  define MRN_HAVE_HA_EXTRA_PREPARE_FOR_FORCED_CLOSE
+#endif
+
 class ha_mroonga;
 
 /* structs */
@@ -107,7 +128,6 @@ struct st_mrn_ft_info
 /* handler class */
 class ha_mroonga: public handler
 {
-  THR_LOCK thr_lock;
   THR_LOCK_DATA thr_lock_data;
 
   MRN_SHARE *share;
@@ -271,12 +291,12 @@ public:
   ha_rows multi_range_read_info_const(uint keyno, RANGE_SEQ_IF *seq,
                                       void *seq_init_param,
                                       uint n_ranges, uint *bufsz,
-                                      uint *flags, COST_VECT *cost);
+                                      uint *flags, Cost_estimate *cost);
   ha_rows multi_range_read_info(uint keyno, uint n_ranges, uint keys,
 #ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ_INFO_KEY_PARTS
                                 uint key_parts,
 #endif
-                                uint *bufsz, uint *flags, COST_VECT *cost);
+                                uint *bufsz, uint *flags, Cost_estimate *cost);
   int multi_range_read_init(RANGE_SEQ_IF *seq, void *seq_init_param,
                             uint n_ranges, uint mode,
                             HANDLER_BUFFER *buf);
@@ -357,6 +377,7 @@ protected:
   int index_last(uchar *buf);
 #endif
   void change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
+  bool primary_key_is_clustered();
 
 private:
   void push_warning_unsupported_spatial_index_search(enum ha_rkey_function flag);
@@ -385,7 +406,7 @@ private:
                      HA_CREATE_INFO *info, MRN_SHARE *tmp_share);
   int storage_create(const char *name, TABLE *table,
                      HA_CREATE_INFO *info, MRN_SHARE *tmp_share);
-  int wrapper_validate_key_info(KEY *key_info);
+  int wrapper_create_index_fulltext_validate(KEY *key_info);
   int wrapper_create_index_fulltext(grn_obj *grn_table,
                                     const char *grn_table_name,
                                     int i,
@@ -416,13 +437,14 @@ private:
   int wrapper_delete_index(const char *name, MRN_SHARE *tmp_share,
                            const char *table_name);
   int storage_delete_table(const char *name, MRN_SHARE *tmp_share,
-                           const char *tbl_name);
+                           const char *table_name);
   int wrapper_open(const char *name, int mode, uint test_if_locked);
   int wrapper_open_indexes(const char *name);
   int storage_open(const char *name, int mode, uint test_if_locked);
   int open_table(const char *name);
   int storage_open_columns(void);
   int storage_open_indexes(const char *name);
+  void wrapper_overwrite_index_bits();
   int wrapper_close();
   int storage_close();
   int mrn_extra(enum ha_extra_function operation);
@@ -552,26 +574,26 @@ private:
                                               uint n_ranges,
                                               uint *bufsz,
                                               uint *flags,
-                                              COST_VECT *cost);
+                                              Cost_estimate *cost);
   ha_rows storage_multi_range_read_info_const(uint keyno,
                                               RANGE_SEQ_IF *seq,
                                               void *seq_init_param,
                                               uint n_ranges,
                                               uint *bufsz,
                                               uint *flags,
-                                              COST_VECT *cost);
+                                              Cost_estimate *cost);
   ha_rows wrapper_multi_range_read_info(uint keyno, uint n_ranges, uint keys,
 #ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ_INFO_KEY_PARTS
                                         uint key_parts,
 #endif
                                         uint *bufsz, uint *flags,
-                                        COST_VECT *cost);
+                                        Cost_estimate *cost);
   ha_rows storage_multi_range_read_info(uint keyno, uint n_ranges, uint keys,
 #ifdef MRN_HANDLER_HAVE_MULTI_RANGE_READ_INFO_KEY_PARTS
                                         uint key_parts,
 #endif
                                         uint *bufsz, uint *flags,
-                                        COST_VECT *cost);
+                                        Cost_estimate *cost);
   int wrapper_multi_range_read_init(RANGE_SEQ_IF *seq, void *seq_init_param,
                                     uint n_ranges, uint mode,
                                     HANDLER_BUFFER *buf);
@@ -622,8 +644,8 @@ private:
                            const char *to_table_name);
   int storage_rename_table(const char *from, const char *to,
                            MRN_SHARE *tmp_share,
-                           const char *from_tbl_name,
-                           const char *to_tbl_name);
+                           const char *from_table_name,
+                           const char *to_table_name);
   bool wrapper_is_crashed() const;
   bool storage_is_crashed() const;
   bool wrapper_auto_repair() const;
@@ -693,6 +715,8 @@ private:
   int storage_start_stmt(THD *thd, thr_lock_type lock_type);
   void wrapper_change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
   void storage_change_table_ptr(TABLE *table_arg, TABLE_SHARE *share_arg);
+  bool wrapper_primary_key_is_clustered();
+  bool storage_primary_key_is_clustered();
 };
 
 #ifdef __cplusplus
