@@ -85,6 +85,26 @@ for test_suite_name in $(echo $test_suite_names | sed -e 's/,/ /g'); do
     fi
 done
 
+innodb_test_suite_dir="${build_test_suites_dir}/innodb"
+groonga_wrapper_innodb_test_suite_dir="${build_test_suites_dir}/groonga_wrapper_innodb"
+if test "$0" -nt "$(dirname "${groonga_wrapper_innodb_test_suite_dir}")"; then
+    rm -rf "${groonga_wrapper_innodb_test_suite_dir}"
+fi
+if ! test -d "${groonga_wrapper_innodb_test_suite_dir}"; then
+    cp -rp "${innodb_test_suite_dir}" "${groonga_wrapper_innodb_test_suite_dir}"
+    ruby -i'' \
+	-pe "\$_.gsub!(/\bengine\s*=\s*innodb\b([^;\\n]*)/i,
+                      \"ENGINE=groonga\\\1 COMMENT='ENGINE \\\"InnoDB\\\"'\")
+            " \
+	${groonga_wrapper_innodb_test_suite_dir}/r/*.result \
+	${groonga_wrapper_innodb_test_suite_dir}/t/*.test \
+	${groonga_wrapper_innodb_test_suite_dir}/include/*.inc \
+	${build_test_include_dir}/innodb_*.inc # over our work. :<
+    sed -i'' \
+	-e '1 i --source include/have_groonga.inc' \
+	${groonga_wrapper_innodb_test_suite_dir}/t/*.test
+fi
+
 if test -n "${plugins_dir}"; then
     make -C ${top_dir} \
 	install-pluginLTLIBRARIES \
